
import numpy as np
from copy import deepcopy
from .hpfitter import HyperparameterFitter

class HyperparameterFitterGPAtom(HyperparameterFitter):
    def __init__(self,func,optimizer=None,bounds=None,use_update_pdis=False,get_prior_mean=False,use_stored_sols=False,**kwargs):
        """ 
        A wrapper for hyperparameter fitter object, so it can be used with ase-GPatom. 
        Hyperparameter fitter object with an optimizer for optimizing the hyperparameters on different given objective functions. 
        Parameters:
            func : ObjectiveFunction class
                A class with the objective function used to optimize the hyperparameters.
            optimizer : Optimizer class
                A class with the used optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                Most of the global optimizers are using boundary conditions. 
                The bounds in this class will be used for the optimizer and func.
                The bounds have to be with the hyperparameter names used in the objective function.
            use_update_pdis : bool
                Whether to update the prior distributions of the hyperparameters with the given boundary conditions.
            get_prior_mean : bool
                Whether to get the parameters of the prior mean in the solution.
            use_stored_sols : bool
                Whether to store the solutions.
        """
        super().__init__(func,
                         optimizer=optimizer,
                         bounds=bounds,
                         use_update_pdis=use_update_pdis,
                         get_prior_mean=get_prior_mean,
                         use_stored_sols=use_stored_sols,
                         **kwargs)
    
    def get_hyperparams(self,hp,model,**kwargs):
        " Get default hyperparameters if they are not given. "
        if hp is None:
            # Get the hyperparameters from the model
            hp=model.hp.copy()
        # Convert to hyperparameter used in the objective function
        hp=self.convert_hp_from_gpatom(hp)
        # Get the values and hyperparameter names
        theta,parameters=self.hp_to_theta(hp)
        return hp,theta,parameters
    
    def update_pdis(self,pdis,model,X,Y,parameters,**kwargs):
        " Update the prior distributions of the hyperparameters with the boundary conditions. "
        pdis=self.convert_dict_object_to_gpatom(pdis)
        return super().update_pdis(pdis,model,X,Y,parameters,**kwargs)
    
    def get_full_hp(self,sol,model,**kwargs):
        " Get the full hyperparameter dictionary with hyperparameters that are optimized and are not. "
        if 'hp' in sol.keys():
            sol['hp']=self.convert_hp_to_gpatom(sol['hp'],model)
        sol['full hp']=model.hp.copy()
        sol['full hp'].update(sol['hp'])
        if 'prefactor' in sol['full hp'].keys():
            sol['full hp'].pop('prefactor')
        sol['full hp']['noise']=sol['full hp']['weight']*sol['full hp']['ratio']
        return sol

    def copy_model(self,model,**kwargs):
        " Copy the model and check if the noisefactor is not used in the factorization method. "
        model=deepcopy(model)
        if 'noisefactor' in model.hp.keys():
            from .objectivefunctions.factorized_likelihood import FactorizedLogLikelihood
            if isinstance(self.func,FactorizedLogLikelihood):
                if model.hp['noisefactor']!=1.0:
                    raise Exception('Noisefactor must be 1.0 for the Factorization method') 
        return model
    
    def convert_hp_from_gpatom(self,hp,**kwargs):
        " Convert the hyperparameters from GP-atom to the form here. "
        parameters=list(hp.keys())
        hp_new={}
        if 'scale' in parameters:
            hp_new['length']=np.array(np.log(hp['scale'])).reshape(-1)
        if 'weight' in parameters:
            hp_new['prefactor']=np.array(np.log(hp['weight'])).reshape(-1)
        if 'ratio' in parameters:
            hp_new['noise']=np.array(np.log(hp['ratio'])).reshape(-1)
        return hp_new
    
    def convert_hp_to_gpatom(self,hp,model,**kwargs):
        " Convert the hyperparameters from here to the form of GP-atom. "
        parameters=list(hp.keys())
        hp_new={}
        if 'length' in parameters:
            hp_new['scale']=np.array(np.exp(hp['length'])).reshape(-1)
        if 'prefactor' in parameters:
            hp_new['weight']=np.array(np.exp(hp['prefactor'])).reshape(-1)[0]
        if 'noise' in parameters:
            hp_new['ratio']=np.array(np.exp(hp['noise'])).reshape(-1)[0]
        return hp_new

    def convert_dict_object_to_gpatom(self,dict_obj,**kwargs):
        " Convert a dictionary with objects with GPatom hyperparameter names to the form here."
        if dict_obj is None:
            return dict_obj
        dict_obj_new={}
        for key,value in dict_obj.items():
            if key=='scale':
                dict_obj_new['length']=dict_obj['scale'].copy()
            if key=='weight':
                dict_obj_new['prefactor']=dict_obj['weight'].copy()
            if key=='ratio':
                dict_obj_new['noise']=dict_obj['ratio'].copy()
            else:
                dict_obj_new[key]=value.copy()
        return dict_obj_new



import numpy as np
from .pdistributions import Prior_distribution
from scipy.special import loggamma

class Invgamma_prior(Prior_distribution):
    def __init__(self,a=1e-20,b=1e-20,**kwargs):
        """ 
        Inverse-Gamma prior distribution used for each type of hyperparameters in log-space. 
        The inverse-gamma distribution is variable transformed from linear- to log-space.
        If the type of the hyperparameter is multi dimensional (H) it is given in the axis=-1. 
        If multiple values (M) of the hyperparameter(/s) are calculated simultaneously it has to be in a (M,H) array. 
        Parameters:
            a: float or (H) array
                The shape parameter. 
            b: float or (H) array
                The scale parameter.
        """
        self.update_arguments(a=a,b=b,**kwargs)
    
    def ln_pdf(self,x):
        if self.nosum:
            return self.lnpre-2.0*self.a*x-self.b*np.exp(-2.0*x)
        return np.sum(self.lnpre-2.0*self.a*x-self.b*np.exp(-2.0*x),axis=-1)
    
    def ln_deriv(self,x):
        return -2.0*self.a+2.0*self.b*np.exp(-2.0*x)
    
    def update_arguments(self,a=None,b=None,**kwargs):
        """
        Update the object with its arguments. The existing arguments are used if they are not given.
        Parameters:
            start: float or (H) array
                The start of non-zero prior distribution value of the hyperparameter in log-space. 
            end: float or (H) array
                The end of non-zero prior distribution value of the hyperparameter in log-space. 
            prob: float or (H) array
                The non-zero prior distribution value.
        Returns:
            self: The updated object itself.
        """
        if a is not None:
            if isinstance(a,(float,int)):
                self.a=a
            else:
                self.a=np.array(a).reshape(-1)
        if b is not None:
            if isinstance(b,(float,int)):
                self.b=b
            else:
                self.b=np.array(b).reshape(-1)
        self.lnpre=np.log(2.0)+self.a*np.log(self.b)-loggamma(self.a)
        if isinstance(self.a,(float,int)) and isinstance(self.b,(float,int)):
            self.nosum=True
        else:
            self.nosum=False
        return self
            
    def mean_var(self,mean,var):
        mean,var=np.exp(mean),np.exp(2.0*np.sqrt(var))
        min_v=mean-np.sqrt(var)*2.0
        return self.update_arguments(a=min_v,b=min_v)
    
    def min_max(self,min_v,max_v):
        b=np.exp(2.0*min_v)
        return self.update_arguments(a=b,b=b)
    
    def copy(self):
        return self.__class__(a=self.a,b=self.b)
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(a=self.a,b=self.b)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs



import numpy as np
from .pdistributions import Prior_distribution

class Gen_normal_prior(Prior_distribution):
    def __init__(self,mu=0.0,s=10.0,v=2,**kwargs):
        """ 
        Independent Generalized Normal prior distribution used for each type of hyperparameters in log-space. 
        If the type of the hyperparameter is multi dimensional (H) it is given in the axis=-1. 
        If multiple values (M) of the hyperparameter(/s) are calculated simultaneously it has to be in a (M,H) array. 
        Parameters:
            mu: float or (H) array
                The mean of the generalized normal distribution. 
            std: float or (H) array
                The scale of the generalized normal distribution.
            v: float or (H) array
                The shape or magnitude of the generalized normal distribution.
        """
        self.update_arguments(mu=mu,
                              s=s,
                              v=v,
                              **kwargs)
    
    def ln_pdf(self,x):
        if self.nosum:
            return -((x-self.mu)/self.s)**(2*self.v)-np.log(self.s)+np.log(0.52)
        return np.sum(-((x-self.mu)/self.s)**(2*self.v)-np.log(self.s)+np.log(0.52),axis=-1)
    
    def ln_deriv(self,x):
        return (-(2.0*self.v)*((x-self.mu)**(2*self.v-1)))/(self.s**(2*self.v))
    
    def update_arguments(self,mu=None,s=None,v=None,**kwargs):
        """
        Update the object with its arguments. The existing arguments are used if they are not given.
        Parameters:
            mu: float or (H) array
                The mean of the generalized normal distribution. 
            std: float or (H) array
                The scale of the generalized normal distribution.
            v: float or (H) array
                The shape or magnitude of the generalized normal distribution.
        Returns:
            self: The updated object itself.
        """
        if mu is not None:
            if isinstance(mu,(float,int)):
                self.mu=mu
            else:
                self.mu=np.array(mu).reshape(-1)
        if s is not None:
            if isinstance(s,(float,int)):
                self.s=s
            else:
                self.s=np.array(s).reshape(-1)
        if v is not None:
            if isinstance(v,(float,int)):
                self.v=v
            else:
                self.v=np.array(v).reshape(-1)
        if isinstance(self.mu,(float,int)) and isinstance(self.s,(float,int)) and isinstance(self.v,(float,int)):
            self.nosum=True
        else:
            self.nosum=False
        return self
            
    def mean_var(self,mean,var):
        return self.update_arguments(mu=mean,s=np.sqrt(var/0.32))
    
    def min_max(self,min_v,max_v):
        mu=(max_v+min_v)/2.0
        return self.update_arguments(mu=mu,s=np.sqrt(2.0/0.32)*(max_v-mu))
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(mu=self.mu,s=self.s,v=self.v)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs



from .localoptimizer import LocalOptimizer
import numpy as np

class LineSearchOptimizer(LocalOptimizer):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=True,theta_index=None,parallel=False,xtol=None,ftol=None,**kwargs):
        """
        The line search optimizer is used for optimzing the objective function wrt. a single hyperparameter.
        The LineSearchOptimizer does only work together with a GlobalOptimizer that uses line searches (e.g. FactorizedOptimizer). 
        A line of the hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set the default theta_index
        self.theta_index=None
        # Set xtol and ftol to the tolerance if they are not given.
        xtol,ftol=self.set_tols(tol,xtol=xtol,ftol=ftol)
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              tol=tol,
                              optimize=optimize,
                              multiple_min=multiple_min,
                              theta_index=theta_index,
                              parallel=parallel,
                              xtol=xtol,
                              ftol=ftol,
                              **kwargs)

    def run(self,func,line,parameters,model,X,Y,pdis,**kwargs):
        """
        Run the line search by optimizing the objective function wrt. the hyperparameter.
        The grid/line of the hyperparameter has to be given.
        Parameters:
            func : ObjectiveFunction class object
                The objective function class that is used to calculate the value.
            line : (ngrid,H) array
                An array with the grid points of the hyperparameters.
                Only one of the hyperparameters is used, which is given by theta_index.
            parameters : (H) list of strings
                A list of names of the hyperparameters.
            model : Model class object
                The Machine Learning Model with kernel and prior that are optimized.
            X : (N,D) array
                Training features with N data points and D dimensions.
            Y : (N,1) array or (N,D+1) array
                Training targets with or without derivatives with N data points.
            pdis : dict
                A dict of prior distributions for each hyperparameter type.
        Returns:
            dict : A solution dictionary with objective function value, optimized hyperparameters,
                success statement, and number of used evaluations.
        """
        raise NotImplementedError()

    def update_arguments(self,maxiter=None,tol=None,optimize=None,multiple_min=None,theta_index=None,parallel=None,xtol=None,ftol=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if tol is not None:
            self.tol=tol
        if optimize is not None:
            self.optimize=optimize
        if multiple_min is not None:
            self.multiple_min=multiple_min
        if theta_index is not None:
            self.theta_index=int(theta_index)
        if parallel is not None:
            self.parallel=parallel
        if xtol is not None:
            self.xtol=xtol
        if ftol is not None:
            self.ftol=ftol
        return self
    
    def find_minimas(self,xvalues,fvalues,i_min,len_l,theta_index,**kwargs):
        " Find all the local minimums and their indicies or just the global minimum and then check convergence. "
        # Investigate multiple minimums
        if self.multiple_min:
            return self.find_multiple_min(xvalues,fvalues,len_l,theta_index)
        # Investigate the global minimum
        return self.find_single_min(xvalues,fvalues,i_min,len_l,theta_index)
    
    def find_multiple_min(self,xvalues,fvalues,len_l,theta_index,**kwargs):
        " Find all the local minimums and their indicies and then check convergence. "
        # Find local minimas for middel part of line 
        i_minimas=np.where((fvalues[1:-1]<fvalues[:-2])&(fvalues[2:]>fvalues[1:-1]))[0]+1
        # Check if the function values for the minimums are within the tolerance
        if len(i_minimas):
            i_keep=abs(fvalues[i_minimas+1]+fvalues[i_minimas-1]-2.0*fvalues[i_minimas])>=self.ftol*(1.0+abs(fvalues[i_minimas]))
            i_minimas=i_minimas[i_keep]
        # Find local minimas for end parts of line
        if fvalues[0]-fvalues[1]<-self.ftol:
            i_minimas=np.append([1],i_minimas)
        if fvalues[-1]-fvalues[-2]<-self.ftol:
            i_minimas=np.append(i_minimas,[len_l-2])
        # Check the distances in the local minimas are within the tolerance
        if len(i_minimas):
            i_keep=abs(xvalues[i_minimas+1,theta_index]-xvalues[i_minimas-1,theta_index])>=self.xtol*(1.0+abs(xvalues[i_minimas,theta_index]))
            i_minimas=i_minimas[i_keep]
        # Sort the indicies after function value sizes
        if len(i_minimas)>1:
            i_sort=np.argsort(fvalues[i_minimas])
            i_minimas=i_minimas[i_sort]
        return i_minimas
    
    def find_single_min(self,xvalues,fvalues,i_min,len_l,theta_index,**kwargs):
        " Find the global minimum and then check convergence. "
        # Investigate the global minimum
        i_minimas=[]
        if i_min==0:
            # Check if the function values are converged if the endpoints is the minimum value
            if fvalues[0]-fvalues[1]<-self.ftol:
                i_minimas=[i_min+1]
        elif i_min==int(len_l-1):
            # Check if the function values are converged if the endpoints is the minimum value
            if fvalues[-1]-fvalues[-2]<-self.ftol:
                i_minimas=[i_min-1]
        else:
            # Check if the function values are converged
            if abs(fvalues[i_min+1]+fvalues[i_min-1]-2.0*fvalues[i_min])>=self.ftol*(1.0+abs(fvalues[i_min])):
                i_minimas=[i_min]
        # Check if the distance in the local minimum is converged
        if len(i_minimas):
            i_minima=i_minimas[0]
            if not abs(xvalues[i_minima+1,theta_index]-xvalues[i_minima-1,theta_index])>=self.xtol*(1.0+abs(xvalues[i_minima,theta_index])):
                i_minimas=[]
        return np.array(i_minimas)
    
    def get_theta_index(self,parameters=[],**kwargs):
        " Get the theta_index. "
        if self.theta_index is None:
            if 'length' in parameters:
                return list(parameters).index('length')
            return 0
        return self.theta_index

    def set_tols(self,tol,xtol=None,ftol=None,**kwargs):
        " Set xtol and ftol to the tolerance if they are not given. "
        if xtol is None:
            xtol=tol
        if ftol is None:
            ftol=tol
        return xtol,ftol

    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,
                        tol=self.tol,
                        optimize=self.optimize,
                        multiple_min=self.multiple_min,
                        theta_index=self.theta_index,
                        parallel=self.parallel,
                        xtol=self.xtol,
                        ftol=self.ftol)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs

    
class GoldenSearch(LineSearchOptimizer):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=True,theta_index=None,parallel=False,xtol=None,ftol=None,**kwargs):
        """
        The golden section search method is used as the line search optimizer.
        The line search optimizer is used for optimzing the objective function wrt. a single the hyperparameter.
        The GoldenSearch does only work together with a GlobalOptimizer that uses line searches (e.g. FactorizedOptimizer). 
        A line of the hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        super().__init__(maxiter=maxiter,
                         tol=tol,
                         optimize=optimize,
                         multiple_min=multiple_min,
                         theta_index=theta_index,
                         parallel=parallel,
                         xtol=xtol,
                         ftol=ftol,
                         **kwargs)

    def run(self,func,line,parameters,model,X,Y,pdis,**kwargs):
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,self.jac,**kwargs)
        # Get the index of the line optimized hyperparameter 
        theta_index=self.get_theta_index(parameters)
        # Calculate function values for line coordinates
        len_l=len(line)
        line=line.reshape(len_l,-1)
        f_list=self.calculate_values(line,func,func_args=func_args)
        # Find the optimal value
        i_min=np.nanargmin(f_list)
        sol={'fun':f_list[i_min],'x':line[i_min],'success':False,'nfev':len_l,'nit':len_l}
        # Check whether the object function is flat
        if (np.nanmax(f_list)-f_list[i_min])<self.ftol:
            i=int(np.floor(0.3*(len(line)-1)))
            return {'fun':f_list[i],'x':line[i],'success':False,'nfev':len_l,'nit':len_l}
        # Find local minimums or the global minimum
        i_minimas=self.find_minimas(line,f_list,i_min,len_l,theta_index)
        # Check for convergence 
        len_i=len(i_minimas)
        if len_i==0:
            sol['success']=True
            return sol
        # Do multiple golden section search if necessary
        if self.optimize and self.maxiter>2:
            sol=self.prepare_run_golden(func,sol,i_minimas,line,f_list,theta_index,func_args=func_args)
        return sol
    
    def prepare_run_golden(self,func,sol,i_minimas,line,f_list,theta_index,func_args=(),**kwargs):
        " Prepare and run the golden section search for the minimums. "
        niter=sol['nfev']
        # Get the function that evaluate the objective function
        fun=self.get_fun(func)
        for i_min in i_minimas:
            # Find the indicies of the interval
            x1=i_min-1
            x4=i_min+1
            # Get the function values of the endpoints of the interval
            f1,f4=f_list[x1],f_list[x4]
            # Get the initial vector as the lower interval coordinates
            theta0=line[x1].copy()
            # Get the direction vector to the upper interval coordinates
            direc=line[x4]-theta0
            # Calculate the norm of the direction vector for the important dimension
            direc_norm=abs(direc[theta_index])
            # Perform the golden section search in the interval
            sol_o=self.golden_search(fun,[0.0,1.0],fbracket=[f1,f4],
                                     maxiter=int(self.maxiter-niter),func_args=func_args,
                                     vec0=theta0,direc=direc,direc_norm=direc_norm)
            # Update the solution
            niter+=sol_o['nfev']
            if sol_o['fun']<=sol['fun']:
                sol=sol_o.copy()
            if niter>=self.maxiter:
                break
        sol['nfev'],sol['nit']=niter,niter
        return sol

    def golden_search(self,fun,bracket,maxiter=200,func_args=(),fbracket=None,vec0=np.array([0.0]),direc=np.array([1.0]),direc_norm=None,**kwargs):
        " Perform a golden section search. "
        # Golden ratio
        r=(np.sqrt(5)-1)/2
        c=1-r
        # Number of function evaluations
        nfev=0
        # Get the coordinates and function values of the endpoints of the interval
        x1,x4=bracket
        vec1=vec0+direc*x1
        vec4=vec0+direc*x4
        if fbracket is None:
            f1,f4=fun(vec1,*func_args),fun(vec4,*func_args)
            nfev+=2
        else:
            f1,f4=fbracket
        # Direction vector norm
        if direc_norm is None:
            direc_norm=np.linalg.norm(direc)
        # Check if the maximum number of iterations have been used
        if maxiter<3:
            i_min=np.nanargmin([f1,f4])
            sol={'fun':[f1,f4][i_min],'x':[vec1,vec4][i_min],'success':False,'nfev':nfev,'nit':nfev}
            return sol
        # Check if the convergence criteria is already met in terms of the coordinates
        if abs(x4-x1)*direc_norm<=self.xtol:
            i_min=np.nanargmin([f1,f4])
            sol={'fun':[f1,f4][i_min],'x':[vec1,vec4][i_min],'success':True,'nfev':nfev,'nit':nfev}
            return sol
        # Make and calculate points within the interval
        x_list=[x1,r*x1+c*x4,c*x1+r*x4,x4]
        f_list=[f1,fun(vec0+direc*x_list[1],*func_args),fun(vec0+direc*x_list[2],*func_args),f4]
        nfev+=2
        # Perform the line search
        success=False
        while nfev<maxiter:
            i_min=np.nanargmin(f_list)
            # Check for convergence
            if np.nanmax(f_list)-f_list[i_min]<=self.ftol*(1.0+abs(f_list[i_min])) or abs(x_list[3]-x_list[0])*direc_norm<=self.xtol*(1.0+direc_norm*abs(x_list[1])):
                success=True
                break
            # Calculate a new point 
            if i_min<2:
                x_list[3]=x_list[2]
                f_list[3]=f_list[2]
                x_list[2]=x_list[1]
                f_list[2]=f_list[1]
                x_list[1]=r*x_list[2]+c*x_list[0]
                f_list[1]=fun(vec0+direc*x_list[1],*func_args)
            else:
                x_list[0]=x_list[1]
                f_list[0]=f_list[1]
                x_list[1]=x_list[2]
                f_list[1]=f_list[2]
                x_list[2]=r*x_list[1]+c*x_list[3]
                f_list[2]=fun(vec0+direc*x_list[2],*func_args)
            nfev+=1
        # Get the solution
        i_min=np.nanargmin(f_list)
        sol={'fun':f_list[i_min],'x':vec0+direc*(x_list[i_min]),'success':success,'nfev':nfev,'nit':nfev}
        return sol


class FineGridSearch(LineSearchOptimizer):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=True,ngrid=80,loops=3,theta_index=None,parallel=False,xtol=None,ftol=None,**kwargs):
        """
        The fine grid search method is used as the line search optimizer.
        The line search optimizer is used for optimzing the objective function wrt. a single the hyperparameter.
        Finer grids are made for all minimums of the objective function.
        The FineGridSearch does only work together with a GlobalOptimizer that uses line searches (e.g. FactorizedOptimizer).
        A line of the hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            loops : int
                The number of loops where the grid points are made.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set the default theta_index
        self.theta_index=None
        # Set xtol and ftol to the tolerance if they are not given.
        xtol,ftol=self.set_tols(tol,xtol=xtol,ftol=ftol)
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              tol=tol,
                              optimize=optimize,
                              multiple_min=multiple_min,
                              ngrid=ngrid,
                              loops=loops,
                              theta_index=theta_index,
                              parallel=parallel,
                              xtol=xtol,
                              ftol=ftol,
                              **kwargs)

    def run(self,func,line,parameters,model,X,Y,pdis,**kwargs):
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,self.jac,**kwargs)
        # Get the index of the line optimized hyperparameter 
        theta_index=self.get_theta_index(parameters)
        # Make empty solution and lists
        sol=self.get_empty_solution()
        lines=np.empty((0,len(line[0])))
        f_lists=np.empty((0))
        # Get the solution from loops of the fine grid method
        sol=self.run_grid_loops(func,sol,line,
                                lines,f_lists,theta_index,
                                loops=self.loops,maxiter=self.maxiter,
                                func_args=func_args)
        return sol
    
    def update_arguments(self,maxiter=None,tol=None,optimize=None,multiple_min=None,ngrid=None,loops=None,theta_index=None,parallel=None,xtol=None,ftol=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            loops : int
                The number of loops where the grid points are made.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if tol is not None:
            self.tol=tol
        if optimize is not None:
            self.optimize=optimize
        if multiple_min is not None:
            self.multiple_min=multiple_min
        if theta_index is not None:
            self.theta_index=int(theta_index)
        if parallel is not None:
            self.parallel=parallel
        if ngrid is not None:
            if self.parallel:
                from ase.parallel import world
                self.ngrid=int(int(ngrid/world.size)*world.size)
                if self.ngrid==0:
                    self.ngrid=world.size
            else:
                self.ngrid=ngrid
        if loops is not None:
            self.loops=int(loops)
        if xtol is not None:
            self.xtol=xtol
        if ftol is not None:
            self.ftol=ftol
        return self

    def run_grid_loops(self,func,sol,line,lines,f_lists,theta_index,loops=3,maxiter=400,func_args=(),**kwargs):
        " Calculate finer grid points and find minimums in a loop if optimize=True. "
        # Calculate function values for line coordinates
        len_l=len(line)
        line=line.reshape(len_l,-1)
        f_list=self.calculate_values(line,func,func_args=func_args)
        # Use previously calculated grid points
        lines=np.append(lines,line,axis=0)
        i_sort=np.argsort(lines[:,theta_index])
        lines=lines[i_sort]
        f_lists=np.append(f_lists,f_list)[i_sort]
        # Find the minimum value
        i_min=np.nanargmin(f_lists)
        # Update the solution dictionary
        sol['nfev']+=len_l
        sol['nit']+=len_l
        if f_lists[i_min]<=sol['fun']:
            sol['fun']=f_lists[i_min]
            sol['x']=lines[i_min]
        # Find local minimums or the global minimum
        i_minimas=self.find_minimas(lines,f_lists,i_min,len(lines),theta_index)
        # Check for convergence
        len_i=len(i_minimas)
        if len_i==0:
            sol['success']=True
            return sol
        # Optimize the minimums
        if self.optimize and loops>0 and maxiter>self.ngrid:
            # Make a new grid if minimums exist
            newline,lines,f_lists=self.make_new_line(lines,f_lists,i_minimas,theta_index,len_i)
            return self.run_grid_loops(func,sol,newline,lines,f_lists,theta_index,loops=int(loops-1),maxiter=int(maxiter-len_l),func_args=func_args)
        return sol

    def make_new_line(self,lines,f_lists,i_minimas,theta_index,len_i,**kwargs):
        " Make a new line/grid for the minimums to optimize the hyperparameter. "
        # Find the grid points that must be saved for later
        i_all=(i_minimas+np.array([[-1],[0],[1]],dtype=int)).T.reshape(-1)
        saved_lines=lines[i_all]
        saved_f_lists=f_lists[i_all]
        # Make a new grid if minimums exist
        if self.multiple_min:
            # If 3 grid points can not be used per minumum then use the lowest minimums
            if self.ngrid<len_i*3:
                i_minimas=i_minimas[:self.ngrid//3]
                len_i=len(i_minimas)
            # Get the number of grid points for each minimum
            di=np.full(shape=len_i,fill_value=self.ngrid//len_i,dtype=int)
            # Get an extra grid point to the lowest minimums if there are grid points to spare
            di[:int(self.ngrid%len_i)]+=1
            # Make new line
            newline=np.concatenate([np.linspace(lines[i-1],lines[i+1],di[j]+2)[1:-1] for j,i in enumerate(i_minimas)])
        else:
            i_min=i_minimas[0]
            # Make new line
            newline=np.linspace(lines[i_min-1],lines[i_min+1],self.ngrid+2)[1:-1]
        return newline,saved_lines,saved_f_lists
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,
                        tol=self.tol,
                        optimize=self.optimize,
                        multiple_min=self.multiple_min,
                        ngrid=self.ngrid,
                        loops=self.loops,
                        theta_index=self.theta_index,
                        parallel=self.parallel,
                        xtol=self.xtol,
                        ftol=self.ftol)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
        

class TransGridSearch(FineGridSearch):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=True,ngrid=80,loops=3,use_likelihood=True,theta_index=None,parallel=False,xtol=None,ftol=None,**kwargs):
        """
        The variable transformed grid search method is used as the line search optimizer.
        The line search optimizer is used for optimzing the objective function wrt. a single the hyperparameter.
        Grids are made by updating the variable transformation from the objective function values.
        The TransGridSearch does only work together with a GlobalOptimizer that uses line searches (e.g. FactorizedOptimizer).
        A line of the hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            loops : int
                The number of loops where the grid points are made.
            use_likelihood : bool
                Whether to use the objective function as a log-likelihood or not.
                If the use_likelihood=False, the objective function is scaled and 
                shifted with the maximum value.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set the default theta_index
        self.theta_index=None
        # Set xtol and ftol to the tolerance if they are not given.
        xtol,ftol=self.set_tols(tol,xtol=xtol,ftol=ftol)
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              tol=tol,
                              optimize=optimize,
                              multiple_min=multiple_min,
                              ngrid=ngrid,
                              loops=loops,
                              use_likelihood=use_likelihood,
                              theta_index=theta_index,
                              parallel=parallel,
                              xtol=xtol,
                              ftol=ftol,
                              **kwargs)
        
    def update_arguments(self,maxiter=None,tol=None,optimize=None,multiple_min=None,ngrid=None,loops=None,use_likelihood=None,theta_index=None,parallel=None,xtol=None,ftol=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            loops : int
                The number of loops where the grid points are made.
            use_likelihood : bool
                Whether to use the objective function as a log-likelihood or not.
                If the use_likelihood=False, the objective function is scaled and 
                shifted with the maximum value.
            theta_index : int or None
                The index of the hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the length-scale.
                If theta_index=None and no length-scale, then theta_index=0.
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if tol is not None:
            self.tol=tol
        if optimize is not None:
            self.optimize=optimize
        if multiple_min is not None:
            self.multiple_min=multiple_min
        if theta_index is not None:
            self.theta_index=int(theta_index)
        if parallel is not None:
            self.parallel=parallel
        if ngrid is not None:
            if self.parallel:
                from ase.parallel import world
                self.ngrid=int(int(ngrid/world.size)*world.size)
                if self.ngrid==0:
                    self.ngrid=world.size
            else:
                self.ngrid=ngrid
        if loops is not None:
            self.loops=int(loops)
        if use_likelihood is not None:
            self.use_likelihood=use_likelihood
        if xtol is not None:
            self.xtol=xtol
        if ftol is not None:
            self.ftol=ftol
        return self

    def make_new_line(self,lines,f_lists,i_minimas,theta_index,len_i,**kwargs):
        " Make new line/grid points from the variable transformation of the objective function. "
        from scipy.integrate import cumulative_trapezoid
        # Change the function to likelihood or to a scaled function from 0 to 1
        if self.use_likelihood:
            fs=np.exp(-(f_lists-np.nanmin(f_lists)))
        else:
            fs=-(f_lists-np.nanmax(f_lists))
            fs=fs/np.nanmax(fs)
        # Calculate the cumulative distribution function values on the grid
        cdf=cumulative_trapezoid(fs,x=lines[:,theta_index],initial=0.0)
        cdf=cdf/cdf[-1]
        cdf_r=cdf.reshape(-1,1)
        # Make new grid points on the inverse cumulative distribution function
        dl=np.finfo(float).eps
        newlines=np.linspace(0.0+dl,1.0-dl,self.ngrid)
        # Find the intervals where the new grid points are located
        i_new=np.where((cdf_r[:-1]<=newlines)&(newlines<cdf_r[1:]))[0]
        i_new_a=i_new+1
        # Calculate the linear interpolation for the intervals of interest
        slope=(lines[i_new_a]-lines[i_new])/(cdf_r[i_new_a]-cdf_r[i_new])
        intercept=((lines[i_new]*cdf_r[i_new_a])-(lines[i_new_a]*cdf_r[i_new]))/(cdf_r[i_new_a]-cdf_r[i_new])
        # Calculate the hyperparameters
        newline=slope*newlines.reshape(-1,1)+intercept
        return newline,lines,f_lists
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,
                        tol=self.tol,
                        optimize=self.optimize,
                        multiple_min=self.multiple_min,
                        ngrid=self.ngrid,
                        loops=self.loops,
                        use_likelihood=self.use_likelihood,
                        theta_index=self.theta_index,
                        parallel=self.parallel,
                        xtol=self.xtol,
                        ftol=self.ftol)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs




import numpy as np
from .length import LengthBoundaries

class RestrictedBoundaries(LengthBoundaries):
    def __init__(self,bounds_dict={},scale=1.0,log=True,max_length=True,use_derivatives=False,**kwargs):
        """ 
        Boundary conditions for the hyperparameters with educated guess for the length-scale and relative-noise hyperparameters.
        Machine precisions are used as default boundary conditions for the rest of the hyperparameters not given in the dictionary.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
            max_length : bool
                Whether to use the maximum scaling for the length-scale or use a more reasonable scaling.
            use_derivatives : bool
                Whether the derivatives of the target are used in the model. 
                The boundary conditions of the length-scale hyperparameter(s) will change with the use_derivatives. 
                The use_derivatives will be updated when update_bounds is called.
        """
        super().__init__(bounds_dict=bounds_dict,scale=scale,log=log,max_length=max_length,use_derivatives=use_derivatives,**kwargs)
    
    def make_bounds(self,model,X,Y,parameters,parameters_set,**kwargs):
        " Make the boundary conditions with educated guesses of the length-scale hyperparamer(s). "
        eps_lower,eps_upper=self.get_boundary_limits()
        self.get_use_derivatives(model)
        bounds={}
        for para in parameters_set:
            if para=='length':
                bounds[para]=self.length_bound(X,parameters.count(para))
            elif para=='noise':
                if 'noise_deriv' in parameters_set:
                    bounds[para]=self.noise_bound(Y[:,0:1],eps_lower=eps_lower)
                else:
                    bounds[para]=self.noise_bound(Y,eps_lower=eps_lower)
            elif para=='noise_deriv':
                bounds[para]=self.noise_bound(Y[:,1:],eps_lower=eps_lower)
            elif para in self.bounds_dict:
                bounds[para]=self.bounds_dict[para].copy()
            else:
                bounds[para]=np.full((parameters.count(para),2),[eps_lower,eps_upper])
        return bounds
    
    def noise_bound(self,Y,eps_lower=10*np.sqrt(2.0*np.finfo(float).eps),**kwargs):
        " Get the minimum and maximum ranges of the noise in the educated guess regime within a scale. "
        n_max=len(Y.reshape(-1))*self.scale
        if self.log:
            return np.array([[eps_lower,np.log(n_max)]])
        return np.array([[eps_lower,n_max]])



import numpy as np
from scipy.linalg import cho_factor,cho_solve
from .objectivefunction import ObjectiveFuction

class ObjectiveFuctionGPAtom(ObjectiveFuction):
    
    def convert_hp_to_gpatom(self,hp,model):
        " Convert the hyperparameters from here to the form of GP-atom. "
        parameters=list(hp.keys())
        hp_new={'weight':1.0}
        if 'length' in parameters:
            hp_new['scale']=np.array(np.exp(hp['length'])).reshape(-1)
        if 'prefactor' in parameters:
            # Prefactor is needed for likelihood function evaluations
            hp_new['prefactor']=np.array(hp['prefactor']).reshape(-1)
        else:
            hp_new['prefactor']=np.array(np.log(model.hp['weight'])).reshape(-1)
        if 'noise' in parameters:
            hp_new['ratio']=np.exp(hp['noise'])[0]
        return hp_new
    
    def get_hyperparams(self,model,**kwargs):
        " Get the hyperparameters for the model and the kernel. "
        return model.hp.copy()
    
    def update_model(self,model,hp,**kwargs):
        " Update the hyperparameters of the machine learning model "
        hp_new=self.convert_hp_to_gpatom(hp,model)
        model.set_hyperparams(hp_new)
        return model
    
    def kxx_corr(self,model,X,**kwargs):
        " Get covariance matrix with or without noise correction. "
        # Calculate the kernel with and without noise
        KXX=model.kernel.kernel_matrix(X)
        n_data=len(KXX)
        KXX=self.add_correction(model,KXX,n_data)
        return KXX,n_data
    
    def kxx_reg(self,model,X,Y,**kwargs):
        " Get covariance matrix with regularization. "
        KXX,n_data=self.kxx_corr(model,X)
        Natoms=int(((len(Y.flatten())/len(X))-1)/3)
        KXX_n=model.add_regularization(KXX.copy(),len(X),Natoms)
        return KXX_n,KXX,n_data
        
    def y_prior(self,X,Y,model,L=None,low=None,**kwargs):
        " Update prior and subtract target. "
        Y_p=Y.copy()
        model.prior.update(X,Y_p,L=L,low=low,**kwargs)
        #self.prior.update(self.X, self.Y, self.L)
        if model.use_forces:
            return (Y_p-model.prior.get(X,Y_p,get_derivatives=True)).reshape(-1,1)
        return (Y_p-model.prior.get(X,Y_p,get_derivatives=False))[:,0:1]
    
    def coef_cholesky(self,model,X,Y,**kwargs):
        " Calculate the coefficients by using Cholesky decomposition. "
        # Calculate the kernel with and without noise
        KXX_n,KXX,n_data=self.kxx_reg(model,X,Y)
        # Cholesky decomposition
        L,low=cho_factor(KXX_n)
        # Subtract the prior mean to the training target
        Y_p=self.y_prior(X,Y,model,L=L,low=low)
        # Get the coefficients
        coef=cho_solve((L,low),Y_p,check_finite=False)
        return coef,L,low,Y_p,KXX,n_data




import numpy as np
from .linesearcher import LineSearchOptimizer,GoldenSearch,FineGridSearch,TransGridSearch


class NoiseGrid(LineSearchOptimizer):
    def __init__(self,maxiter=5000,**kwargs):
        """
        The grid method is used as the line search optimizer.
        The grid of relative-noise hyperparameter values is calculated with the objective function.
        The lowest of objective function values of the single grid is used as the optimum.
        A line of the relative-noise hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,**kwargs)
        
    def run(self,func,line,parameters,model,X,Y,pdis,**kwargs):
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,self.jac,**kwargs)
        # Calculate function values for line coordinates
        len_l=len(line)
        line=line.reshape(len_l,-1)
        f_list=self.calculate_values(line,func,func_args=func_args)
        # Find the optimal value
        i_min=np.nanargmin(f_list)
        sol={'fun':f_list[i_min],'x':line[i_min],'success':False,'nfev':len_l,'nit':len_l}
        return sol

    def update_arguments(self,maxiter=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        return self
        
    def get_func_arguments(self,parameters,model,X,Y,pdis,jac,func_args=(),**kwargs):
        " Get the arguments needed for the objective function. "
        return func_args
    
    def get_fun(self,func,**kwargs):
        " Get the function that evaluates the objective function. "
        return func.get_eig_fun
    
    def calculate_values(self,thetas,func,func_args=(),**kwargs):
        " Calculate a list of values with a function. "
        return func.get_all_eig_fun(thetas,*func_args)

    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs


class NoiseGoldenSearch(GoldenSearch):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=False,theta_index=0,xtol=None,ftol=None,**kwargs):
        """
        The golden section search method is used as the line search optimizer.
        The line search optimizer is used for optimzing the objective function wrt. the relative-noise hyperparameter.
        A line of the relative-noise hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            theta_index : int or None
                The index of the relative-noise hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the relative-noise.
                If theta_index=None and no relative-noise, then theta_index=0.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set the default theta_index
        self.theta_index=None
        # Set xtol and ftol to the tolerance if they are not given.
        xtol,ftol=self.set_tols(tol,xtol=xtol,ftol=ftol)
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              tol=tol,
                              optimize=optimize,
                              multiple_min=multiple_min,
                              theta_index=theta_index,
                              xtol=xtol,
                              ftol=ftol,
                              **kwargs)
    
    def get_func_arguments(self,parameters,model,X,Y,pdis,jac,func_args=(),**kwargs):
        " Get the arguments needed for the objective function. "
        return func_args
    
    def get_fun(self,func,**kwargs):
        " Get the function that evaluates the objective function. "
        return func.get_eig_fun
    
    def calculate_values(self,thetas,func,func_args=(),**kwargs):
        " Calculate a list of values with a function. "
        return func.get_all_eig_fun(thetas,*func_args)


class NoiseFineGridSearch(FineGridSearch):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=False,ngrid=80,loops=2,theta_index=0,xtol=None,ftol=None,**kwargs):
        """
        The fine grid search method is used as the line search optimizer.
        The line search optimizer is used for optimzing the objective function wrt. the relative-noise hyperparameter.
        Finer grids are made for all minimums of the objective function.
        A line of the relative-noise hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            loops : int
                The number of loops where the grid points are made.
            theta_index : int or None
                The index of the relative-noise hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the relative-noise.
                If theta_index=None and no relative-noise, then theta_index=0.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set the default theta_index
        self.theta_index=None
        # Set xtol and ftol to the tolerance if they are not given.
        xtol,ftol=self.set_tols(tol,xtol=xtol,ftol=ftol)
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              tol=tol,
                              optimize=optimize,
                              multiple_min=multiple_min,
                              ngrid=ngrid,
                              loops=loops,
                              theta_index=theta_index,
                              xtol=xtol,
                              ftol=ftol,
                              **kwargs)
    
    def get_func_arguments(self,parameters,model,X,Y,pdis,jac,func_args=(),**kwargs):
        " Get the arguments needed for the objective function. "
        return func_args
    
    def get_fun(self,func,**kwargs):
        " Get the function that evaluates the objective function. "
        return func.get_eig_fun
    
    def calculate_values(self,thetas,func,func_args=(),**kwargs):
        " Calculate a list of values with a function. "
        return func.get_all_eig_fun(thetas,*func_args)


class NoiseTransGridSearch(TransGridSearch):
    def __init__(self,maxiter=5000,tol=1e-5,optimize=True,multiple_min=False,ngrid=80,loops=2,use_likelihood=True,theta_index=0,xtol=None,ftol=None,**kwargs):
        """
        The variable transformed grid search method is used as the line search optimizer.
        The line search optimizer is used for optimzing the objective function wrt. the relative-noise hyperparameter.
        Grids are made by updating the variable transformation from the objective function values.
        A line of the relative-noise hyperparameter is required to run the line search.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            tol : float
                A tolerance criterion for convergence. 
            optimize : bool
                Whether to optimize the line given by split it into smaller intervals. 
            multiple_min : bool
                Whether to optimize multiple minimums or just optimize the lowest minimum.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            loops : int
                The number of loops where the grid points are made.
            use_likelihood : bool
                Whether to use the objective function as a log-likelihood or not.
                If the use_likelihood=False, the objective function is scaled and 
                shifted with the maximum value.
            theta_index : int or None
                The index of the relative-noise hyperparameter that is optimized with the line search.
                If theta_index=None, then it will use the index of the relative-noise.
                If theta_index=None and no relative-noise, then theta_index=0.
            xtol : float
                A tolerance criterion of the hyperparameter for convergence. 
            ftol : float
                A tolerance criterion of the objective function for convergence. 
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Line search optimizers cannot use gradients of the objective function 
        self.jac=False
        # Set the default theta_index
        self.theta_index=None
        # Set xtol and ftol to the tolerance if they are not given.
        xtol,ftol=self.set_tols(tol,xtol=xtol,ftol=ftol)
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              tol=tol,
                              optimize=optimize,
                              multiple_min=multiple_min,
                              ngrid=ngrid,
                              loops=loops,
                              use_likelihood=use_likelihood,
                              theta_index=theta_index,
                              xtol=xtol,
                              ftol=ftol,
                              **kwargs)
        
    def get_func_arguments(self,parameters,model,X,Y,pdis,jac,func_args=(),**kwargs):
        " Get the arguments needed for the objective function. "
        return func_args
    
    def get_fun(self,func,**kwargs):
        " Get the function that evaluates the objective function. "
        return func.get_eig_fun
    
    def calculate_values(self,thetas,func,func_args=(),**kwargs):
        " Calculate a list of values with a function. "
        return func.get_all_eig_fun(thetas,*func_args)
    



import numpy as np
from .objectivefunction_gpatom import ObjectiveFuctionGPAtom

class FactorizedLogLikelihood(ObjectiveFuctionGPAtom):
    def __init__(self,get_prior_mean=False,modification=False,ngrid=80,bounds=None,noise_optimizer=None,**kwargs):
        """ 
        The factorized log-likelihood objective function that is used to optimize the hyperparameters. 
        The prefactor hyperparameter is determined from an analytical expression. 
        An eigendecomposition is performed to get the eigenvalues. 
        The relative-noise hyperparameter can be searched from a single eigendecomposition for each length-scale hyperparameter.  
        Parameters:
            get_prior_mean: bool
                Whether to save the parameters of the prior mean in the solution.
            modification: bool
                Whether to modify the analytical prefactor value in the end.
                The prefactor hyperparameter becomes larger if modification=True.
            ngrid: int
                Number of grid points that are searched in the relative-noise hyperparameter. 
            bounds: Boundary_conditions class
                A class of the boundary conditions of the relative-noise hyperparameter.
            noise_optimizer : Noise line search optimizer class
                A line search optimization method for the relative-noise hyperparameter.
        """
        # Set descriptor of the objective function
        self.use_analytic_prefactor=True
        self.use_optimized_noise=True
        # Set default bounds
        if bounds is None:
            from ..hpboundary.hptrans import VariableTransformation
            bounds=VariableTransformation(bounds=None)
        # Set default noise line optimizer
        if noise_optimizer is None:
            from ..optimizers.noisesearcher import NoiseFineGridSearch
            noise_optimizer=NoiseFineGridSearch(maxiter=1000,tol=1e-5,optimize=True,multiple_min=False,ngrid=ngrid,loops=2)
        # Set the arguments
        self.update_arguments(get_prior_mean=get_prior_mean,
                              modification=modification,
                              ngrid=ngrid,
                              bounds=bounds,
                              noise_optimizer=noise_optimizer,
                              **kwargs)

    def update_arguments(self,get_prior_mean=None,modification=None,ngrid=None,bounds=None,noise_optimizer=None,**kwargs):
        """
        Update the objective function with its arguments. The existing arguments are used if they are not given.
        Parameters:
            get_prior_mean: bool
                Whether to save the parameters of the prior mean in the solution.
            modification: bool
                Whether to modify the analytical prefactor value in the end.
                The prefactor hyperparameter becomes larger if modification=True.
            ngrid: int
                Number of grid points that are searched in the relative-noise hyperparameter. 
            bounds: Boundary_conditions class
                A class of the boundary conditions of the relative-noise hyperparameter.
            noise_optimizer : Noise line search optimizer class
                A line search optimization method for the relative-noise hyperparameter.
        Returns:
            self: The updated object itself.
        """
        if get_prior_mean is not None:
            self.get_prior_mean=get_prior_mean
        if modification is not None:
            self.modification=modification
        if ngrid is not None:
            self.ngrid=int(ngrid)
        if bounds is not None:
            self.bounds=bounds.copy()
        if noise_optimizer is not None:
            self.noise_optimizer=noise_optimizer.copy()
        # Always reset the solution when the objective function is changed 
        self.reset_solution()
        return self

    def function(self,theta,parameters,model,X,Y,pdis=None,jac=False,**kwargs):
        hp,parameters_set=self.make_hp(theta,parameters)
        model=self.update_model(model,hp)
        D,U,Y_p,UTY,KXX,n_data=self.get_eig(model,X,Y)
        noise,nlp=self.maximize_noise(parameters,model,X,Y,pdis,hp,UTY,D,n_data)
        if jac:
            deriv=self.derivative(hp,parameters_set,model,X,KXX,D,U,Y_p,UTY,noise,pdis,**kwargs)
            self.update_solution(nlp,theta,hp,model,jac=jac,deriv=deriv,noise=noise,UTY=UTY,D=D,n_data=n_data)
            return nlp,deriv
        self.update_solution(nlp,theta,hp,model,jac=jac,noise=noise,UTY=UTY,D=D,n_data=n_data)
        return nlp
    
    def derivative(self,hp,parameters_set,model,X,KXX,D,U,Y_p,UTY,noise,pdis,**kwargs):
        nlp_deriv=np.array([])
        D_n=D+np.exp(2.0*noise)
        prefactor2=np.mean(UTY/D_n)
        hp['prefactor']=np.array([0.5*np.log(prefactor2)])
        hp['noise']=np.array([noise])
        KXX_inv=np.matmul(U/D_n,U.T)
        coef=np.matmul(KXX_inv,Y_p)
        for para in parameters_set:
            if para=='prefactor':
                nlp_deriv=np.append(nlp_deriv,np.zeros((len(hp[para]))))
                continue
            K_deriv=self.get_K_deriv(model,para,X=X,KXX=KXX)
            K_deriv_cho=self.get_K_inv_deriv(K_deriv,KXX_inv)
            nlp_deriv=np.append(nlp_deriv,(-(0.5*np.matmul(coef.T,np.matmul(K_deriv,coef)).reshape(-1))/prefactor2)+0.5*K_deriv_cho)
        nlp_deriv=nlp_deriv-self.logpriors(hp,pdis,jac=True)
        return nlp_deriv
    
    def get_eig_fun(self,noise,hp,pdis,UTY,D,n_data,**kwargs):
        " Calculate log-likelihood from Eigendecomposition for a noise value. "
        D_n=D+np.exp(2.0*noise)
        prefactor=0.5*np.log(np.mean(UTY/D_n))
        nlp=0.5*n_data*(1+np.log(2.0*np.pi))+(n_data*prefactor)+0.5*np.sum(np.log(D_n))
        if pdis is not None:
            hp['prefactor']=np.array([prefactor])
            hp['noise']=np.array([noise]).reshape(-1)
        return nlp-self.logpriors(hp,pdis,jac=False)
    
    def get_all_eig_fun(self,noises,hp,pdis,UTY,D,n_data,**kwargs):
        " Calculate log-likelihood from Eigendecompositions for all noise values from the list. "
        D_n=D+np.exp(2.0*noises)
        prefactor=0.5*np.log(np.mean(UTY/D_n,axis=1))
        nlp=(0.5*n_data*(1+np.log(2.0*np.pi)))+((n_data*prefactor)+(0.5*np.sum(np.log(D_n),axis=1)))
        if pdis is not None:
            hp['prefactor']=prefactor.reshape(-1,1)
            hp['noise']=noises
        return nlp-self.logpriors(hp,pdis,jac=False)
    
    def make_noise_list(self,model,X,Y,**kwargs):
        " Make the list of noises. " 
        return self.bounds.make_single_line(parameter='noise',ngrid=self.ngrid).reshape(-1,1)
    
    def maximize_noise(self,parameters,model,X,Y,pdis,hp,UTY,D,n_data,**kwargs):
        " Find the maximum relative-noise with a grid method. "
        noises=self.make_noise_list(model,X,Y)
        # Make the function arguments
        func_args=(hp.copy(),pdis,UTY,D,n_data)
        # Calculate function values for line coordinates
        sol=self.noise_optimizer.run(self,noises,['noise'],model,X,Y,pdis,func_args=func_args)
        # Find the minimum value
        return sol['x'][0],sol['fun']

    def update_solution(self,fun,theta,hp,model,jac=False,deriv=None,noise=None,UTY=None,D=None,n_data=None,**kwargs):
        """
        Update the solution of the optimization in terms of hyperparameters and model.
        The lowest objective function value is stored togeher with its hyperparameters.
        The prior mean can also be saved if get_prior_mean=True.
        The prefactor and relative-noise hyperparameters are stored as different values
        than the input since they are optimized analytically and numerically, respectively.
        """
        if fun<self.sol['fun']:
            D_n=D+np.exp(2.0*noise)
            prefactor2=np.mean(UTY/D_n)
            if self.modification:
                prefactor2=(n_data/(n_data-len(theta)))*prefactor2 if n_data-len(theta)>0 else prefactor2
            hp['prefactor']=np.array([0.5*np.log(prefactor2)])
            hp['noise']=np.array([noise])            
            self.sol['x']=np.concatenate([hp[para] for para in sorted(hp.keys())])
            self.sol['hp']=hp.copy()
            self.sol['fun']=fun
            if jac:
                self.sol['jac']=deriv.copy()
            if self.get_prior_mean:
                self.sol['prior']=self.get_prior_parameters(model)
        return self.sol
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(get_prior_mean=self.get_prior_mean,
                        modification=self.modification,
                        ngrid=self.ngrid,
                        bounds=self.bounds,
                        noise_optimizer=self.noise_optimizer)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs




import numpy as np
from .prior import Prior

class Prior_constant(Prior):
    def __init__(self,yp=0.0,add=0.0,**kwargs):
        """
        The prior mean of the targets. 
        The prior mean is used as a baseline of the target values.
        The prior mean is a constant from the target values if given else it is 0. 
        A value can be added to the constant.
        Parameters:
            yp : float
                The prior mean constant
            add : float
                A value added to the found prior mean from data.
        """
        self.update_arguments(yp=yp,add=add,**kwargs)

    def get(self,X,Y,get_derivatives=True,**kwargs):
        if get_derivatives:
            yp=np.zeros(Y.shape)
            yp[:,0]=self.yp
            return yp
        return np.full(Y.shape,self.yp)
    
    def get_parameters(self,**kwargs):
        return dict(yp=self.yp,add=self.add)
    
    def update_arguments(self,yp=None,add=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            yp : float
                The prior mean constant
            add : float
                A value added to the found prior mean from data.
        Returns:
            self: The updated object itself.
        """
        if add is not None:
            self.add=add
        if yp is not None:
            self.yp=yp+self.add
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(yp=self.yp,add=self.add)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
